/****************************************************  Utility Functions  *****************************************************

	Red Rapids
	797 North Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions are specific to the Signal Stream application example.

	Rev History:  07-14-2015:	Modified the MemTest() function to exercise the new VHDL model that replaced the MIG core.
				  10-20-2014:	Modified the PCI throughput benchmark to activate all available DMA channels.
				  01-10-2014:	Inital release.

	Copyright 2014, Red Rapids, Inc.

****************************************************  Utility Functions  *****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "RRadapter0.h"
#include "sigfpga.h"
#include "dma_functions.h"
#include "common_functions.h"
#include "channel_functions.h"

// Clear the PCI latency performance measurement.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ClearLatency(s_Adapter *p_HW)
{
	unsigned long Status = 0;

	// Clear the PCI latency performance register.
	Status += Write(p_HW, BAR2, PCI_LATENCY, 0x0ull);

	return(Status);
}

// Preload a transmitter buffer with sine wave data samples.
// Return of zero indicates that the buffer contains an integer number of sine wave periods, non-zero indicates a fractional number of periods.
unsigned long DataLoad(void **p_PageAddresses, s_Channel TXChannel, double FsDivisor, size_t DataItemSize)
{
	unsigned long long max = 0xFFFFFFFFFFFFFFFFull;
	size_t page_count = 0;
	size_t record_count = 0;
	size_t records_per_page = 0;
	size_t data_items_per_record = 0;
	size_t data_items_count = 0;
	unsigned long long *virtual_address;
	double phase = 0;
	double phase_select = 0;
	double data_item_size_dbl = 0;

	// Cast the DataItemSize variable for use with other data types.
	data_item_size_dbl = (double)DataItemSize; 

	// Compute the number of 64-bit records per DMA page.
	records_per_page = TXChannel.DMABuffer.BurstSizeB * TXChannel.DMABuffer.BurstCount / 8;

	// Compute the number of data items per 64-bit record.
	data_items_per_record = (size_t)floor(64 / data_item_size_dbl);

	// Sequence through the number of DMA pages per mark since this one chapter will be used for every DMA buffer refresh.
	for(page_count=0; page_count < TXChannel.DMABuffer.PagesPerMark; page_count++)
	{
		// Get the virtual address of the first 64-bit record in the current page.
		virtual_address = (unsigned long long *)p_PageAddresses[page_count];

		// Sequence through every 64-bit record in the page.
		for(record_count = 0; record_count < records_per_page; record_count++)
		{
			// Clear the current 64-bit record.
			*(virtual_address + record_count) = 0x0ull;
			
			for(data_items_count = 1; data_items_count <= data_items_per_record; data_items_count++)
			{
				// Compute the amplitude of the DAC sample based on phase and write it as the current data item to the payload record.
				*(virtual_address + record_count) |= (((max >> (64-DataItemSize)) & (unsigned long long)((pow(2,data_item_size_dbl-1)-1)*sin(2*3.141592654*phase/FsDivisor))) << (64-(data_items_count*DataItemSize)));

				// Compute the phase of the next sample.
				phase = fmod(++phase_select,FsDivisor);
				// Restart the phase select counter if the phase is zero.
				if (!phase) phase_select = 0;
			}
		}
	}
	if (phase_select)
		return 1;
	else
		return 0;
}

// Write the contents of the requested number of DMA pages associated with the requested channel number to the specified file.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long DataSave(void **p_PageAddresses, s_Channel RXChannel, char *p_File, size_t PageCount, size_t DataItemSize)
{
	unsigned long Status = 0;
	FILE *fp;
	unsigned long long one = 0x1, max = 0xFFFFFFFFFFFFFFFFull;
	size_t page_count = 0;
	size_t data_items_count = 0;
	size_t record_count = 0;
	size_t records_per_page = 0;
	signed sample = 0;
	unsigned long long *virtual_address;

	// Open the specified data file.
	fp = fopen(p_File, "w");

	// Process samples only if the file open was successful.
	if(fp != NULL)
	{
		// Compute the number of 64-bit records per DMA page.
		records_per_page = RXChannel.DMABuffer.BurstSizeB * RXChannel.DMABuffer.BurstCount / 8;

		// Sequence through all of the DMA pages assigned to the channel.
		for(page_count=0; page_count < PageCount; page_count++)
		{
			// Get the virtual address of the first 64-bit record in each page.
			virtual_address = (unsigned long long *)p_PageAddresses[page_count];

			// Sequence through every 64-bit record in the page.
			for(record_count = 0; record_count < records_per_page; record_count++)

				// Sequence through every sample in the 64-bit record.
				for (data_items_count = 0; data_items_count < (64 / DataItemSize); data_items_count++)
					{
						// Extract the next sample value from the 64-bit record.
						sample = (signed)(*(virtual_address + record_count) >> (data_items_count * DataItemSize)) & (max >> (64 - DataItemSize));
						// Sign extend negative numbers.
						if (sample & (one << (DataItemSize - 1))) sample |= (max << DataItemSize);
						// Print the sample value to the file.
						fprintf(fp, "%d\n", sample);
					}
		}
		fclose(fp);
		Status = 0;
	}
	else
		Status = 1;

	return(Status);
}

// Report the response time for the last interrupt that was serviced.
// Returns the response time in increments of 4 ns to a maximum value of (2^28)-1, or 1.074 sec.
// Note: The interrupt timeout (see below) will be reported if shorter than the actual response time.
unsigned long InterruptResponse(s_Adapter *p_HW)
{
	unsigned long long Value = 0;

	Read(p_HW, BAR0, INTERRUPT_LATENCY, &Value);

	return((unsigned long)Value);
}

// Set the interrupt timeout in increments of 4 ns to a maximum value of (2^28)-1, or 1.074 sec.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long InterruptTimeout(s_Adapter *p_HW, unsigned long IntTimeout)
{
	return(Write(p_HW, BAR0, INTERRUPT_TIMEOUT, (unsigned long long)IntTimeout));
}

// Report the size of the QDR II+ SRAM address.
// Returns the number of bits.
unsigned long MemSize(s_Adapter *p_HW)
{
	unsigned long long Value = 0;

	Read(p_HW, BAR2, IDENTIFICATION, &Value);

	return((unsigned long)((Value >> 52) & 0x00000000000000FFull));
}

// Command the device to perform a test of every available address in the selected SRAM.
// SRAMSelect = 1:  SRAM-A
// SRAMSelect = 2:  SRAM-B
// SRAMSelect = 3:  SRAM-A and SRAM-B simultaneously
// Return of zero indicates success, non-zero indicates the number of errors recorded.
unsigned long MemTest(s_Adapter *p_HW, unsigned long SRAMSelect)
{
	unsigned long Status = 0;
	unsigned long long Value;
	size_t watchdog = 0;


	// Start the test on the selected SRAM.
	Write(p_HW, BAR2, SRAM_CMD_STAT, SRAMSelect);

	// Wait for the test to complete by monitoring the busy status.
	do
	{
		// Pause while the firmware exercises every address in the SRAM.
		MicrosecondTimer(15000);
		// Read the busy status register.
		Read(p_HW, BAR2, SRAM_CMD_STAT, &Value);
		// Increment the watchdog counter to avoid getting caught in this loop.
		watchdog++;
	}
	while (Value && watchdog != 10);

	// Read the error count register.
	Read(p_HW, BAR2, SRAM_ERROR, &Value);

	// Return a unique value if the watchdog timer expired.
	if (watchdog == 10)
		Status = 1000000000;
	// Return the recorded error count for SRAM-A if selected.
	else if (SRAMSelect == 1)
		Status = Value & 0xFFFFFFFFull;
	// Return the recorded error count for SRAM-B if selected.
	else if (SRAMSelect == 2)
		Status = (Value >> 32) & 0xFFFFFFFFull;
	// Return the sum of the recorded error counts for SRAM-A and SRAM-B if selected.
	else if (SRAMSelect == 3)
		Status = (Value & 0xFFFFFFFFull) + ((Value >> 32) & 0xFFFFFFFFull);

	return(Status);
}

// Benchmark the maximum attainable PCI performance of RX and/or TX DMA transactions.
// WARNING: DMA buffers must be allocated prior to executing this function or the computer will likely crash!
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long PciBenchmark(s_Adapter *p_HW)
{
	unsigned long long Value = 0;
	unsigned long long int_mask = 0;
	unsigned long Status = 0;
	unsigned long busy = 0;

	// Record the current state of the global interrupt mask.
	Status += Read(p_HW, BAR0, GLOBAL_INT_MASK, &int_mask);
	// Disable interrupts during the benchmark.
	Status += Write(p_HW, BAR0, GLOBAL_INT_MASK, 0x0);

	// Initiate continuous DMA transfers on all channels.
	Status += Write(p_HW, BAR2, PCI_THROUGHPUT, 0x1FFull);

	// Pause while DMA transactions are counted for a quarter second.
	MicrosecondTimer(300000);

	// Discontinue continuous DMA transfers on the channels that participated in the benchmark.
	Status += Write(p_HW, BAR2, PCI_THROUGHPUT, 0x0ull);

	// Monitor the status of all channels with a polling loop and terminate when all channels stop processing.
	do
	{
		// Check the current operating state of each channel that participated in the benchmark to determine if any are still active.
		Read(p_HW, BAR0, DMA_1TO2_STATUS, &Value);
		busy = (Value | (Value >> 32)) & 0x00005000ull;
		Read(p_HW, BAR0, DMA_3TO4_STATUS, &Value);
		busy |= (Value | (Value >> 32)) & 0x00005000ull;
		Read(p_HW, BAR0, DMA_5TO6_STATUS, &Value);
		busy |= (Value | (Value >> 32)) & 0x00005000ull;
		Read(p_HW, BAR0, DMA_7TO8_STATUS, &Value);
		busy |= (Value | (Value >> 32)) & 0x00005000ull;
	}
	while (busy);

	// Restore the original interrupt mask state.
	Status += Write(p_HW, BAR0, GLOBAL_INT_MASK, int_mask);

	return(Status);
}

// Report the measured instantaneous PCI throughput for the last 250 ms and the maximum DMA latency since the last clear.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long PciPerformance(s_Adapter *p_HW, unsigned long *p_Performance)
{
	unsigned long long Value = 0;
	unsigned long Status = 0;

	// Read the register that stores PCI throughput every 250 ms.
	Status += Read(p_HW, BAR2, PCI_THROUGHPUT, &Value);

	// Record the measured receiver throughput in bytes/sec.
	p_Performance[0] = Value & 0x00000000FFFFFFFFull;
	// Record the measured transmitter throughput in bytes/sec.
	p_Performance[1] = (Value >> 32) & 0x00000000FFFFFFFFull;

	// Read the register that stores maximum PCI latency.
	Status += Read(p_HW, BAR2, PCI_LATENCY, &Value);

	// Record the measured receiver latency as the number of 4 ns clock periods.
	p_Performance[2] = Value & 0x00000000FFFFFFFFull;
	// Record the measured transmitter latency as the number of 4 ns clock periods.
	p_Performance[3] = (Value >> 32) & 0x00000000FFFFFFFFull;

	return(Status);
}

// Query each channel for any status change and update the user defined structure of the hardware handle accordingly.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ProcessChannels(s_Adapter *p_HW)
{
	unsigned long Status = 0;
	unsigned long CurrentMarker = 0;	// Variable to store the most marker number of the current DMA chapter.
	unsigned long chancnt = 0;			// Variable to index channel numbers in loops.
	unsigned long pagecnt = 0;			// Counter to sequence through the DMA pages of a chapter.
	s_User *User = NULL;				// Pointer to the user defined structure of the hardware handle.
	s_ChannelStatus ChannelStatus;		// Structure to store the results returned by a channel status query to the device.
	s_DMAStatus DMAStatus;				// Structure to store the results returned by a DMA status query to the device.
	s_GlobalStatus GlobalStatus;		// Structure to store the results returned by an interrupt status query to the device.

	// Cast the user defined pointer from the hardware handle.
	User = (s_User *)(p_HW->User);

	// Determine which channels have a status update.
	Status = GetGlobalStatus(p_HW, &GlobalStatus);

	// Update the user defined structure of any channel reporting a status change.
	for (chancnt = 1; chancnt <= User->RXChannelCount + User->TXChannelCount; chancnt++)
	{
		if (GlobalStatus.Channel[chancnt-1])
		{
			// Determine which channel status events require an update.
			Status += GetChannelStatus(p_HW, chancnt, &ChannelStatus);

			// Increment the converter (ADC/DAC) error count if an error is detected.
			if (ChannelStatus.Error[0]) User->Error[chancnt-1][0]++;

			// Increment the FIFO error count if an error is detected.
			if (ChannelStatus.Error[1]) User->Error[chancnt-1][1]++;

			// Process the DMA buffer contents if the DMA refresh marker is detected.
			if (ChannelStatus.DMAMarker)
			{
				// Copy the contents of the last chapter of DMA pages that were refreshed between DMA markers into data buffers accessible to the main application.
				// Note: This assumes that there was only one DMA marker update since the last check.
				for (pagecnt = 0; pagecnt < p_HW->DMA.PagesPerMark[chancnt-1]; pagecnt++)
					// Determine whether the current channel is a transmitter or receiver.
					if (chancnt > User->RXChannelCount)
						// Copy from the data buffer to the DMA buffer if this is a transmitter channel.
						memcpy(p_HW->DMA.VirtDMAAdr[chancnt-1][(User->MarkerIndex[chancnt-1]*p_HW->DMA.PagesPerMark[chancnt-1])+pagecnt], User->p_DataBuffer[chancnt-1][pagecnt], p_HW->DMA.BurstSizeB[chancnt-1]*p_HW->DMA.BurstCount[chancnt-1]);
					else
						// Copy from the DMA buffer to the data buffer if this is a receiver channel.
						memcpy(User->p_DataBuffer[chancnt-1][pagecnt], p_HW->DMA.VirtDMAAdr[chancnt-1][(User->MarkerIndex[chancnt-1]*p_HW->DMA.PagesPerMark[chancnt-1])+pagecnt], p_HW->DMA.BurstSizeB[chancnt-1]*p_HW->DMA.BurstCount[chancnt-1]);

				// Determine which channel status events require an update.
				Status += GetDMAStatus(p_HW, chancnt, &DMAStatus);
				// Compute the current marker index based on the current page of the DMA engine.
				CurrentMarker = (unsigned long)floor(DMAStatus.CurrentPage/p_HW->DMA.PagesPerMark[chancnt-1]);

				// Determine whether the current marker is beyond the next expected marker, which means that data is already lost.
				// There is a roll over condition to consider when the last DMA page is completed and the index starts back at zero.
				// Note: This technique cannot detect loss of data if the DMA engine completes a full cycle through all of the available pages before the test is performed.
				if ((CurrentMarker > User->MarkerIndex[chancnt-1] + 1) || (CurrentMarker == 0 && User->MarkerIndex[chancnt-1] > 0 && User->MarkerIndex[chancnt-1] < User->MaxMarkerIndex[chancnt-1]))
					// Increment the DMA error count if an error is detected.
					User->Error[chancnt-1][2]++;

				// Set the marker index to the current marker position.
				User->MarkerIndex[chancnt-1] = CurrentMarker;
				
			}
		}
	}

	return(Status);
}

// Initialize the structure members assigned to the user defined pointer attached to the hardware handle.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long UserInitialize(s_User *p_User, unsigned long ChannelNumber, s_DMABuffer DMABuffer, unsigned long *p_DataBuffer)
{
	unsigned long Status = 0;
	size_t index = 0;
	
	// Initialize the DMA marker index to the first set of pages.
	p_User->MarkerIndex[ChannelNumber-1] = 0;
	// Compute the maximum marker index based on the DMA buffer settings.
	p_User->MaxMarkerIndex[ChannelNumber-1] = (DMABuffer.PageCount / DMABuffer.PagesPerMark) - 1;
	for (index = 0; index < DMABuffer.PagesPerMark; index++)
	{
		p_User->p_DataBuffer[ChannelNumber-1][index] = malloc(DMABuffer.BurstSizeB * DMABuffer.BurstCount);
		if (p_User->p_DataBuffer[ChannelNumber-1][index] == NULL)
		{
			// Do not continue because there is no place to store data.
			Status++;
			// Mark this data buffer as a failed allocation.
			p_DataBuffer[index] = 1;
		}
		else
			// Mark this data buffer as a successful allocation.
			p_DataBuffer[index] = 0;
	}
	// Reset the error counters.
	for (index = 0; index < 3; index++)
		p_User->Error[ChannelNumber-1][index] = 0;

	return(Status);
}